//
//  Mound3DView.java
//  Termites
//
//  Created by John Montgomery on Wed Feb 11 2004.
//  Copyright (c) 2004 Little Spikey Land. All rights reserved.
//

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import java.util.*;

/** A simple 3D view.  This view creates a static 3D model of the 
 *  mound, allowing the user to get a better feel for the construction.
 *  It uses a voxel-based rendering, with the voxel colours signifying the 
 *  underlying position of an element in the mound, to give better sense of
 *  depth.
 **/
public class Mound3DView extends JPanel {
    private double theta = 0.0f;
    private double cos_theta = Math.cos( theta ), sin_theta = Math.sin( theta );
    private Voxel[] voxels = null;
    private int minX, maxX;
    private int minY, maxY;
    
    private int dragX = 0;
    private double dragTheta = theta;

    private int SIZE = 4;

	/** Make a 3D model of the simulation.  The model is a snapshot in time
	 *  and will not be altered if the sim updates.
	 *  The model can be rotated by the user around the z-axis.
	 **/
    public Mound3DView( TermiteSim sim ) {
        ArrayList voxelList = new ArrayList();
        for ( int i = 0; i < sim.getWidth(); i++ ) {
            for ( int j = 0; j < sim.getHeight(); j++ ) {
                for ( int k = 0; k < sim.getDepth(); k++ ) {
                    int cell = sim.getCell( i, j, k );
                    if ( cell > 0 ) {
                        Color color = new Color( (255*i)/sim.getWidth(), (255*j)/sim.getHeight(), (255*k)/sim.getDepth() );
                        voxelList.add( new Voxel( color, i-sim.getWidth()/2, j-sim.getHeight()/2, k ) );
                    }
                }
            }
        }
        voxels = (Voxel[])voxelList.toArray( new Voxel[ voxelList.size() ] );
        transform();
        minmax();
        
        addMouseMotionListener(
            new MouseMotionAdapter() {
                public void mouseDragged( MouseEvent me ) {
                    int x = (me.getX() - dragX);
                    theta = dragTheta + ((x-getWidth())*Math.PI*2)/getWidth();
                    
                    transform();
                    repaint();
                }
            }
        );
        
        addMouseListener(
            new MouseAdapter() {
                public void mousePressed( MouseEvent me ) {
                    dragX = me.getX();
                    dragTheta = theta;
                    requestFocus();
                }
            }
        );
        
        addKeyListener(
            new KeyAdapter() {
                public void keyReleased( KeyEvent ke ) {
                    int code = ke.getKeyCode();
                    switch( code ) {
                        case KeyEvent.VK_I:  SIZE++; break;
                        case KeyEvent.VK_O: SIZE--; break;
                        default:
                            return;
                    }
                    transform();
                    repaint();
                }
            }
        );

    }
    
    public Dimension getPreferredSize() {
        return new Dimension( maxX-minX, maxY-minY );
    }
    
    private synchronized void minmax() {
        minX = minY = Integer.MAX_VALUE;
        maxX = maxY = 0;
        for ( int i = 0; i < voxels.length; i++ ) {
            voxels[ i ].minmax();
        }
    }
    
    private synchronized void transform() {
        cos_theta = Math.cos( theta );
        sin_theta = Math.sin( theta );
        for ( int i = 0; i < voxels.length; i++ ) {
            voxels[ i ].transform();
        }
        Arrays.sort( voxels );
    }

    public synchronized void paint( Graphics g ) {
        int w = maxX-minX;
        int h = maxY-minY;
        g.clearRect( 0, 0, getWidth() , getHeight() );
        g.translate( -minX - (w-getWidth())/2, -minY - (h-getHeight())/2 );
        for ( int i = 0; i < voxels.length; i++ ) {
            voxels[ i ].paint( g );
        }
    }
    
	/** Simple voxel class.  It has a colour, size and position and can be
	 *  rotated around the z-axis.
	 **/
    private class Voxel implements Comparable {
        private Color color = null;
        private int voxelSize = 5;
        private int x, y, z;
        private int tx, ty, tz;
        private int px, py;
        
        public Voxel( Color color, int x, int y, int z ) {
            this.color = color;
            this.x = x;
            this.y = y;
            this.z = z;
            transform();
        }
        
        public void minmax() {
            minX = Math.min( minX, px );
            maxX = Math.max( maxX, px );
            minY = Math.min( minY, py );
            maxY = Math.max( maxY, py );
        }
        
        public void transform() {
            tx = (int)(SIZE*(x*cos_theta + y*sin_theta));
            ty = (int)(SIZE*(y*cos_theta - x*sin_theta));
            tz = SIZE*z;
            px = (tx-ty)/2;
            py = (ty+tx)/4 - tz;
        }
        
        public void paint( Graphics g ) {
            g.setColor( color );
            g.fillRect( px-SIZE, py-SIZE, SIZE<<1, SIZE<<1 );
        }
        
        public int compareTo( Object o ) {
            Voxel v = (Voxel)o;
            if ( v.tz > tz )
                return -1;
            if ( v.tz < tz )
                return 1;
                
            // tz must be equals
            if ( v.py > py )
                return -1;
            if ( v.py < py )
                return 1;
                
            return 0;
        }
    }

}
