//
//  Termite.java
//  Termites
//
//  Created by John Montgomery on Wed Feb 04 2004.
//  Copyright (c) 2004 Little Spikey Land. All rights reserved.
//

/** The class representing a termite agent in the simulation.
 **/
public class Termite {
    private int[] directions = new int[ 27 ];
    {
        for ( int i = 0; i < directions.length; i++ ) directions[ i ] = i;
    }
    private float[] weights = new float[ 27 ];
    private int x, y, z;
    private boolean building = true;
    private float q = 0.3f;
    private float heightBias = 2000.0f;
    private float pheremoneBias = 1.0f;
    
	/** Make a new termite, that is initially building and place it at (x,y,z).
	 **/
    public Termite( int x, int y, int z ) {
        this.x = x;
        this.y = y;
        this.z = z;
    }
    
    public float getGreediness() {
        return q;
    }
    
    public void setGreediness( float greediness ) {
        this.q = greediness;
    }
    
    public float getHeightBias() {
        return heightBias;
    }
    
    public void setHeightBias( float heightBias ) {
        this.heightBias = heightBias;
    }
    
    public float getPheremoneBias() {
        return pheremoneBias;
    }
    
    public void setPheremoneBias( float pheremoneBias ) {
        this.pheremoneBias = pheremoneBias;
    }
    
    public int getX() {
        return x;
    }
    
    public int getY() {
        return y;
    }
    
    public int getZ() {
        return z;
    }
    
	/** Current state (building or foraging).
	 **/
    public boolean isBuilding() {
        return building;
    }
    
	/** Update this termite, within the given simulation.
	 **/
    public void update( TermiteSim sim ) {
        
        x = Math.max( 0, Math.min( x, sim.getWidth() ) );
        y = Math.max( 0, Math.min( y, sim.getHeight() ) );
        z = Math.max( 0, Math.min( z, sim.getDepth() ) );
        
        if ( !sim.isSupported( x, y, z ) ) {
            do {
                z = (z = z-1) < 0? 0 : z;
            }
            while ( !sim.isSupported( x, y, z ) );
        }
        else if ( sim.getCell( x, y, z ) > 0 ) {
            z = z+1;
        }
        else {
            if ( building ) {
                build( sim );
            }
            else {
                collect( sim );
            }
        }
    }
    
    private void shuffle( int[] arr ) {
        for ( int i = 0; i < arr.length; i++ ) {
            int val = arr[ i ];
            int index = i + (int)(Math.random()*(arr.length-i));
            arr[ i ] = arr[ index ];
            arr[ index ] = val;
        }
    }
    
    private void calculateWeights( TermiteSim sim ) {
        int w = 0;
        for ( int k = -1; k <= 1; k++ ) {
            for ( int j = -1; j <= 1; j++ ) {
                for ( int i = -1; i <= 1; i++, w++ ) {
                    boolean supported = sim.isSupported( x+i, y+j, z+k ) && sim.getCell( x+i, y+j, z+k ) == 0;
                    if ( !supported ) 
                        weights[ w ] = 0.0f;
                    else {
                        int pheremone = sim.getPheremone( x+i, y+j, z+k )+1;
                        weights[ w ] = heightBias*(k+1) + pheremoneBias*pheremone;
                    }
                }
            }
        }
    }
    
    private int chooseDirection() {
        shuffle( directions );
        float sum = 0.0f;
        float bestWeight = -Float.MAX_VALUE;
        int best = -1;
        
        for ( int i = 0; i < weights.length; i++ ) {
            int index = directions[ i ];
            float weight = weights[ index ];
            sum += weight;
            if ( weight > bestWeight ) {
                bestWeight = weight;
                best = index;
            }
        }
        
        if ( Math.random() < q )
            return best;
        
        float threshold = (float)(Math.random()*sum);
        sum = 0.0f;
        for ( int i = 0; i < weights.length; i++ ) {
            float weight = weights[ i ];
            sum += weight;
            if ( sum >= threshold )
                return i;
        }
        
        return 3*(3 +1) + 1;
    }
    
    private void move( int dir, TermiteSim sim ) {
        int dz = dir/9;
        int dy = (dir - (dz*9))/3;
        int dx = dir - (9*dz + 3*dy);
        
        dz--; dy--; dx--;
        
        if ( building ) {
            if ( dx == 0 && dy == 0 && dz == 0 && Math.random() < 0.75 ) {
                building = false;
                sim.dropPheremone( x, y, z );
                sim.setCell( 1, x, y, z );
            }
        }
        
        if ( sim.getCell( x+dx, y+dy, z+dz ) == 0 ) {
            this.z += dz;
            this.y += dy;
            this.x += dx;
        }
    }
    
    private void build( TermiteSim sim ) {
        calculateWeights( sim );
        int dir = chooseDirection();
        move( dir, sim );
    }
    
    private void collect( TermiteSim sim ) {
        if ( z <= 0 )
            building = true;
        else {
            move( (int)(Math.random()*weights.length), sim );
        }
    }
    
    public String toString() {
        return getClass().getName() + "[" + x + "," + y + "," + z + "," + (building? "building" : "collecting" ) + "]";
    }
    
}
