//
//  TermiteApplet.java
//  Termites
//
//  Created by John Montgomery on Fri Feb 06 2004.
//  Copyright (c) 2004 Little Spikey Land. All rights reserved.
//

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.border.*;
import java.util.*;


/** The overall gui class.  This class is responsible for creating the various GUI components and
 *  laying them out, as well as hooking them up to a termite sim.  In effect it is the "controller"
 *  in the "model, view, controller" paradigm.
 **/
public class TermiteApplet extends JApplet implements Runnable, ActionListener, ItemListener, FocusListener  {
    private TermiteSim sim = null;
    private SplitViews view = new SplitViews();
    
    
    private JCheckBox showTermitesCheck = new JCheckBox( "Show Termites", true );
    private JCheckBox showPheremonesCheck = new JCheckBox( "Show Pheremones", false );
    
    private boolean paused = false;
    private JButton stopStartButton = new JButton( "Pause" );
    private JButton newButton       = new JButton( "New" );
    private JButton threeDButton       = new JButton( "3D" );
    
	private JTextField numTermitesText = null;
	
    private JTextField greedinessText = null;
    private JTextField heightBiasText = null;
    private JTextField pheremoneBiasText = null;
    
    private JTextField phRadiusText  = null;
    private JTextField phDecayText   = null;
    
    private Thread thread = null;
    
	/** Layout the components.
	 **/
    public void init() {
        getContentPane().setLayout( new BorderLayout() );
        getContentPane().add( view, BorderLayout.CENTER );
        
        JPanel controls = new JPanel();
        controls.setLayout( new BoxLayout( controls, BoxLayout.Y_AXIS ) );
        
        JPanel buttonPanel = new JPanel();
        
        buttonPanel.add( stopStartButton );
        buttonPanel.add( newButton );
        buttonPanel.add( threeDButton );
        
        controls.add( showTermitesCheck );
        showTermitesCheck.addItemListener( this );
        controls.add( showPheremonesCheck );
        showPheremonesCheck.addItemListener( this );
        
        stopStartButton.addActionListener( this );
        newButton.addActionListener( this );
        threeDButton.addActionListener( this );
                
        controls.add( buttonPanel );
        
        // termite controls
        JPanel termiteControls = new JPanel();
        Border termiteBorder = BorderFactory.createTitledBorder( "Termites" );
        termiteControls.setBorder( termiteBorder );
        
        termiteControls.setLayout( new GridLayout( 4, 2, 10, 10 ) );
	
		numTermitesText = new JTextField( "100" );
		
        greedinessText = new JTextField( "0.3" );
        greedinessText.addActionListener( this );
        greedinessText.addFocusListener( this );
        
        heightBiasText = new JTextField( "2000" );
        heightBiasText.addActionListener( this );
        heightBiasText.addFocusListener( this );
        
        pheremoneBiasText = new JTextField( "0.1" );
        pheremoneBiasText.addActionListener( this );
        pheremoneBiasText.addFocusListener( this );
        
		termiteControls.add( new JLabel( "Termites:" ) );
        termiteControls.add( numTermitesText );
        termiteControls.add( new JLabel( "Greediness:" ) );
        termiteControls.add( greedinessText );
        termiteControls.add( new JLabel( "Height Bias:" ) );
        termiteControls.add( heightBiasText );
        termiteControls.add( new JLabel( "Pheremone Bias:" ) );
        termiteControls.add( pheremoneBiasText );
        
        
        controls.add( termiteControls );
        
        // pheremone controls
        JPanel pheremoneControls = new JPanel();
        Border pheremoneBorder = BorderFactory.createTitledBorder( "Pheremones" );
        pheremoneControls.setBorder( pheremoneBorder );
        
        pheremoneControls.setLayout( new GridLayout( 2, 2, 10, 10 ) );
        
        phRadiusText = new JTextField( "10" );
        phRadiusText.addActionListener( this );
        phRadiusText.addFocusListener( this );

        
        phDecayText = new JTextField( "0.01" );
        phDecayText.addActionListener( this );
        phDecayText.addFocusListener( this );

        pheremoneControls.add( new JLabel( "Radius:" ) );
        pheremoneControls.add( phRadiusText );
        pheremoneControls.add( new JLabel( "Decay Rate:" ) );
        pheremoneControls.add( phDecayText );
        
        controls.add( pheremoneControls );
        
        JPanel eastPanel = new JPanel();
        eastPanel.add( controls );
        
        getContentPane().add( eastPanel, BorderLayout.EAST );
    }
    
    private float getFloatValue( JTextField textField, float min, float max ) {
        String text = textField.getText();
        float val = min;
        try {
            val = new Float( text ).floatValue();
            if ( min <= val && val <= max )
                return val;
            
            if ( val < min )
                val = min;
            
            if ( val > max )
                val = max;
        }
        catch( NumberFormatException nfe ) {
        }
        
        Toolkit.getDefaultToolkit().beep();
        textField.setText( String.valueOf( val ) );
        
        return val;
    }
    
    private void setFromParams() {
        //System.out.println( "setFromParams" );
        
        float greediness = getFloatValue( greedinessText, 0.0f, 1.0f );
        float heightBias = getFloatValue( heightBiasText, 0.0f, Float.MAX_VALUE );
        float pheremoneBias = getFloatValue( pheremoneBiasText, 0.0f, Float.MAX_VALUE );
        
        float decayRate = getFloatValue( phDecayText, 0.0f, 1.0f );
        float radius = getFloatValue( phRadiusText, 0.0f, Float.MAX_VALUE );
        
        if ( sim != null ) {
            sim.setTermiteGreediness( greediness );
            sim.setTermiteHeightBias( heightBias );
            sim.setTermitePheremoneBias( pheremoneBias );
        
            sim.setPheremoneDecayRate( decayRate );
            sim.setPheremoneRadius( (int)radius );
        }
    }
    
    private void setFromSim() {
        //System.out.println( "setFromSim" );
    
        if ( sim != null ) {
            greedinessText.setText( String.valueOf( sim.getTermiteGreediness() ) );
            heightBiasText.setText( String.valueOf( sim.getTermiteHeightBias() ) );
            pheremoneBiasText.setText( String.valueOf( sim.getTermitePheremoneBias() ) );
        
            phDecayText.setText( String.valueOf( sim.getPheremoneDecayRate() ) );
            phRadiusText.setText( String.valueOf( sim.getPheremoneRadius() ) );
        }
    }

    
    public void actionPerformed( ActionEvent ae ) {
        if ( ae.getSource() == newButton ) {
            start();
        }
        else if ( ae.getSource() == stopStartButton ) {
            if ( stopStartButton.getText().equals( "Play" ) ) {
                stopStartButton.setText( "Pause" );
                paused = false;
            }
            else {
                stopStartButton.setText( "Play" );
                paused = true;
            }
        }
        else if ( ae.getSource() == threeDButton ) {
            synchronized( sim ) {
                Mound3DView view = new Mound3DView( sim );
                JFrame frame = new JFrame( "3D View" ); 
                frame.getContentPane().add( view );
                frame.pack();
                frame.show();
            }
        }
        else {
            setFromParams();
        }
    }
    
    public void focusGained( FocusEvent fe ) {}
    
    public void focusLost( FocusEvent fe ) {
        setFromSim();
    }
    
    public void itemStateChanged( ItemEvent ie ) {
        if ( ie.getSource() == showTermitesCheck ) {
            view.setShowTermites( showTermitesCheck.isSelected() );
        }
        else if ( ie.getSource() == showPheremonesCheck ) {
            view.setShowPheremones( showPheremonesCheck.isSelected() );
        }
    }
    
	/** Create a new TermiteSim and spawn a thread to
	 *  run it.
	 **/
    public void start() {
		int numTermites = (int)getFloatValue( numTermitesText, 0, 1001 );
		if ( numTermites > 0 && numTermites <= 1000 ) {
			sim = new TermiteSim( numTermites, 64, 64, 128 );
			view.setTermiteSim( sim );
			setFromParams();
			(thread = new Thread( this )).start();
		}
		else {
			Toolkit.getDefaultToolkit().beep();
			numTermitesText.setText( "100" );
			numTermitesText.requestFocus();
		}
    }

    public void stop() {
        thread = null;
    }
    
	/** This is the main loop that runs the termite sim
	 *  and spawns repaint events, to redraw the views.
	 **/
    public void run() {
        Thread.currentThread().setPriority( Thread.MIN_PRIORITY );
        while( Thread.currentThread() == thread ) {
            if ( !paused ) {
                sim.update();
                view.repaint();
            }
            try {
                Thread.sleep( 40 );
            }
            catch( Exception e ){}
        }
    }

}
