//
//  TermiteSim.java
//  Termites
//
//  Created by John Montgomery on Sat Jan 31 2004.
//  Copyright (c) 2004 Little Spikey Land. All rights reserved.
//

import java.util.List;
import java.util.ArrayList;
import java.util.Random;

/** The class that is the simulation.  This class corresponds to the
 *  "model" in the "model, view, controller" paradigm.
 **/
public class TermiteSim {
    private int width, height, depth;
    private int[] cells = null;
    private int[] pheremones = null;
    private float decayRate = 0.01f;
    private int spread = 10;
    
    private int highest = 0;
    
    private List termites = new ArrayList();
    
    private Random rand = new Random();
    
	/** Created a new simulation with the specified dimensions and
	 *  number of termites.  The termites are distributed randomly
	 *  upon the ground and are initially ready to build.
	 *  A small amount of pheremone is also dropped in the middle of the ground
	 *  to help start the building process.
	 **/
    public TermiteSim( int numTermites, int width, int height, int depth ) {
        this.width  = width;
        this.height = height;
        this.depth  = depth;
        cells      = new int[ width*height*depth ];
        pheremones = new int[ width*height*depth ];
        
        for ( int i = 0; i < numTermites; i++ ) {
            int x = (int)((width/2) + (0.5-Math.random())*width);
            int y = (int)((height/2) + (0.5-Math.random())*height);
            termites.add( new Termite( x, y, 0 ) );
        }
        
        for ( int i = 0; i < 5; i++ )
            dropPheremone( (int)((width/2) + (0.5-Math.random())*15), (int)((height/2) + (0.5-Math.random())*15), 0 );
    }
    
    public float getPheremoneDecayRate() {
        return decayRate;
    }
    
    public synchronized void setPheremoneDecayRate( float decayRate ) {
        this.decayRate = decayRate;
    }
    
    public int getPheremoneRadius() {
        return spread;
    }
    
    public synchronized void setPheremoneRadius( int radius ) {
        this.spread = radius;
    }
    
    public float getTermiteGreediness() {
        float greediness = 0.0f;
        if ( termites.size() > 0 ) {
            Termite termite = (Termite)termites.get( 0 );
            greediness = termite.getGreediness();
        }
        return greediness;
    }

    
    public synchronized void setTermiteGreediness( float greediness ) {
        for ( int i = 0; i < termites.size(); i++ ) {
            Termite termite = (Termite)termites.get( i );
            termite.setGreediness( greediness );
        }
    }

    public float getTermiteHeightBias() {
        float bias = 0.0f;
        if ( termites.size() > 0 ) {
            Termite termite = (Termite)termites.get( 0 );
            bias = termite.getHeightBias();
        }
        return bias;
    }

    
    public synchronized void setTermiteHeightBias( float bias ) {
        for ( int i = 0; i < termites.size(); i++ ) {
            Termite termite = (Termite)termites.get( i );
            termite.setHeightBias( bias );
        }
    }

    public float getTermitePheremoneBias() {
        float bias = 0.0f;
        if ( termites.size() > 0 ) {
            Termite termite = (Termite)termites.get( 0 );
            bias = termite.getPheremoneBias();
        }
        return bias;
    }

    
    public synchronized void setTermitePheremoneBias( float bias ) {
        for ( int i = 0; i < termites.size(); i++ ) {
            Termite termite = (Termite)termites.get( i );
            termite.setPheremoneBias( bias );
        }
    }
    
    
    private int randInt() {
        return Math.abs( rand.nextInt() );
    }
    
    public int getNumTermites() {
        return termites.size();
    }
    
    public Termite getTermite( int i ) {
        return (Termite)termites.get( i );
    }
    
    public int getWidth() {
        return width;
    }
    
    public int getHeight() {
        return height;
    }
    
    public int getDepth() {
        return depth;
    }
    
    public int getHighest() {
        return highest;
    }
    
    public synchronized void update() {
        for ( int i = 0; i < getNumTermites(); i++ ) {
            Termite termite = getTermite( i );
            termite.update( this );
        }
        updatePheremones();
    }
    
    public int getCell( int x, int y, int z ) {
        if ( x >= 0 && x < width && y >= 0 && y < height && z >= 0 && z < depth ) {
            return cells[ index( x, y, z ) ];
        }
        if ( z < 0 ) return 1; // bottom is solid
        
        return 0;
    }
    
    public void setCell( int value, int x, int y, int z ) {
        if ( x >= 0 && x < width && y >= 0 && y < height && z >= 0 && z < depth ) {
            cells[ index( x, y, z ) ] = value;
            highest = Math.max( z, highest );
            //System.out.println( "Set " + x + "," + y + "," + z );
        }
    }
    
    public int getPheremone( int x, int y, int z ) {
        if ( x >= 0 && x < width && y >= 0 && y < height && z >= 0 && z < depth ) {
            return pheremones[ index( x, y, z ) ];
        }
        return 0;
    }
    
    private int index( int x, int y, int z ) {
        return width*(height*z + y) + x;
    }
    
    public void dropPheremone( int x, int y, int z ) {
        //System.out.println( "Dropped at: " + x + "," + y + "," + z );
        //add_pheremone( 128, x, y, z );
		if ( spread == 0 ) return;
        int r = spread;
        int d = (int)( 3*r*r );
        for ( int i = -r; i <= r; i++ ) {
            for ( int j = -r; j <= r; j++ ) {
                for ( int k = -r; k <= r; k++ ) {
                    int sq = (int)(i*i + j*j + k*k);
                    int p = 1+ (255*(d - sq))/d;
                    //int p = 255;
                    add_pheremone( p, x+i, y+j, z+k );
                }
            }
        }
    }
    
    private void add_pheremone( int amount, int x, int y, int z ) {
        if ( x >= 0 && x < width && y >= 0 && y < height && z >= 0 && z < depth ) {
            pheremones[ index( x, y, z ) ] += amount;
        }
    }
    
    public void sumPheremones( int[] layer, boolean overhead ) {
        if ( overhead ) {
            for ( int z = 0; z < depth; z++ ) {
                for ( int y = 0; y < height; y++ ) {
                    int yWidth = y*width;
                    for ( int x = 0; x < width; x++ ) {
                        layer[ yWidth + x ] += pheremones[ index( x, y, z ) ];
                    }
                }
            }
        }
        else {
            for ( int y = 0; y < height; y++ ) {
                for ( int z = 0; z < depth; z++ ) {
                    for ( int x = 0; x < width; x++ ) {
                        layer[ (depth-z-1)*width + x ] += pheremones[ index( x, y, z ) ];
                    }
                }
            }
        }
    }

	/** Slightly misnamed, but this class finds the "nearest" 
	 *  solid item when viewed from overhead or sideon.
	 **/
    public void sumCells( int[] layer, boolean overhead ) {
        
        if ( overhead ) {
            for ( int z = 0; z < depth; z++ ) {
                for ( int y = 0; y < height; y++ ) {
                    int yWidth = y*width;
                    for ( int x = 0; x < width; x++ ) {
                        if ( cells[ index( x, y, z ) ] > 0 )
                            layer[ yWidth + x ] = z+1;
                    }
		}
            }
        }
	else {
            for ( int y = 0; y < height; y++ ) {
                for ( int z = 0; z < depth; z++ ) {
                    for ( int x = 0; x < width; x++ ) {
			if ( cells[ index( x, y, z ) ] > 0 )
			layer[ (depth-z-1)*width + x ] = y+1;
		    }
		}
            }
        }
    }

    
    public void updatePheremones() {
        int scale = 1024;
        int rate = (int)((1.0f-decayRate)*scale);
        for ( int i = 0; i < pheremones.length; i++ ) pheremones[ i ] = (rate*pheremones[ i ])/scale;
    }
    
	/** Query whether a termite at (x,y,z) would having something to
	 *  hold onto.  i.e. one of the neighbouring elements is not empty.
	 **/
    public boolean isSupported( int x, int y, int z ) {
        for ( int i = -1; i <= 1; i++ ) {
            for ( int j = -1; j <= 1; j++ ) {
                for ( int k = -1; k <= 1; k++ ) {
                    if ( getCell( x+i, y+j, z+k ) > 0 ) return true;
                }
            }
        }
        return false;
    }
    
}
