//
//  TermiteView.java
//  Termites
//
//  Created by John Montgomery on Sat Jan 31 2004.
//  Copyright (c) 2004 Little Spikey Land. All rights reserved.
//

import javax.swing.JPanel;
import java.awt.Graphics;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Container;
import java.awt.image.BufferedImage;
import java.awt.event.*;

/** The TermiteView class is responsible for drawing either side-on or top-down
 *  two dimensional views of the TermiteSim.
 *  It can be set to either draw the "physical" view (where the blocks are) or else the 
 *  pheremone view.  Drawing of the termites can also be disabled, so that the structure 
 *  can be viewed on its own.
 **/
public class TermiteView extends JPanel {
    private TermiteSim sim = null;
    private BufferedImage pheremoneImage = null;
    private int[] pheremonePixels = null;
    private BufferedImage cellImage = null;
    private int[] cellPixels = null;
    private int[] cellDepths = null;
    
    private boolean overhead = true;
    
    private boolean showCells = true;
    private boolean showTermites = true;
    
	/** Creating a new TermiteView, specifying whether it is a
	 *  view from over-head (true) or from side-on (false).
	 **/
    public TermiteView( boolean overhead ) {
		this.overhead = overhead;
	}
    
    private synchronized void changeShowCells() {
        showCells = !showCells;
        repaint();
    }
    
    public synchronized void setShowPheremones( boolean showPheremones ) {
        this.showCells = !showPheremones;
        repaint();
    }
    
    public synchronized void setShowTermites( boolean showTermites ) {
        this.showTermites = showTermites;
        repaint();
    }
    
    public synchronized void setTermiteSim( TermiteSim sim ) {
		int w =-1, h = -1;
		if ( overhead ) {
			w = sim.getWidth();
			h = sim.getHeight();
		}
		else {
			w = sim.getWidth();
			h = sim.getDepth();
		}
        this.sim = sim;
        pheremoneImage = new BufferedImage( w, h, BufferedImage.TYPE_INT_RGB );
        pheremonePixels = new int[ w*h ];
        cellImage = new BufferedImage( w, h, BufferedImage.TYPE_INT_RGB );
        cellPixels = new int[ w*h ];
        cellDepths = new int[ w*h ];
        invalidate();
        Container parent = getParent();
        if ( parent != null )
            parent.doLayout();
    }
    
    public Dimension getPreferredSize() {
        if ( sim == null ) return new Dimension( 128, 128 );
        
        if ( overhead )
            return new Dimension( sim.getWidth(), sim.getHeight() );
        else
            return new Dimension( sim.getWidth(), sim.getDepth() );
    }
    
    private void drawTermites( int[] zbuffer, int[] pixels, int width, int height ) {
        if ( showTermites ) {
            for ( int i = 0; i < sim.getNumTermites(); i++ ) {
                Termite termite = sim.getTermite( i );
                
                int color = termite.isBuilding()? 0xFF0000 : 0x00FF00;
                
                int x = -1, y = -1, z = -1;
                
                if ( overhead ) {
                    x = termite.getX();
                    y = termite.getY();
                    z = termite.getZ()+1;
                }
                else {
                    x = termite.getX();
                    y = sim.getDepth()-termite.getZ();
                    z = termite.getY()+1;
                }
                
                if ( x >= 0 && x < width && y >= 0 && y < height ) {
                    if ( zbuffer[ y*width +x ] <= z )
                        pixels[ y*width +x ] = color;
                }
            }
        }

    }
    
    public synchronized void paint( Graphics g ) {
        g.setColor( Color.white );
        g.fillRect( 0, 0, getWidth(), getHeight() );
        
        if ( sim == null ) return;
        
        int w =-1, h = -1, d = -1;
        if ( overhead ) {
            w = sim.getWidth();
            h = sim.getHeight();
            d = sim.getDepth();
        }
        else {
            w = sim.getWidth();
            h = sim.getDepth();
            d = sim.getHeight();
        }
		
        synchronized( sim ) {
            for ( int i = 0; i < cellDepths.length; i++ ) cellDepths[ i ] = 0;
            sim.sumCells( cellDepths, overhead );
        
            if ( !showCells ) {
                for ( int i = 0; i < pheremonePixels.length; i++ ) pheremonePixels[ i ] = 0;
                sim.sumPheremones( pheremonePixels, overhead );
                int max = 1;
                for ( int i = 0; i < pheremonePixels.length; i++ ) {
                    max = Math.max( pheremonePixels[ i ], max );
                }
                for ( int i = 0; i < pheremonePixels.length; i++ ) {
                    pheremonePixels[ i ] = (255*pheremonePixels[ i ])/max;
                }
                
                drawTermites( cellDepths, pheremonePixels, w, h );
                
                pheremoneImage.setRGB( 0, 0, w, h, pheremonePixels, 0, sim.getWidth() );
                
                g.drawImage( pheremoneImage, 0, 0, getWidth(), getHeight(), this );
            }
            else {
                for ( int i = 0; i < cellPixels.length; i++ ) {
                    int cells = cellDepths[ i ];
                    int color = 0;
                    if ( cells == 0 ) color = 0xFFFFFF;
                    else {
                        int grey = (255*cells)/d;
                        color = (grey << 16) | (grey << 8) | (0x00);
                    }
                    cellPixels[ i ] = color;
                }
                
                drawTermites( cellDepths, cellPixels, w, h );
                
                cellImage.setRGB( 0, 0, w, h, cellPixels, 0, sim.getWidth() );
                g.drawImage( cellImage, 0, 0, getWidth(), getHeight(), this );
            }
            
            //g.setColor( Color.gray );
            //g.drawString( String.valueOf( sim.getTermite( 0 ) ), 5, 20 );
            //g.drawString( "Height: " +  sim.getHighest(), 5, 40 );
            
        }
    }


}
